/* rawio.c: measure raw disk I/O throughput */
/*-
 * Copyright (c) 1999
 *	Nan Yang Computer Services Limited.  All rights reserved.
 *
 *  This software is distributed under the so-called ``Berkeley
 *  License'':
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Nan Yang Computer
 *      Services Limited.
 * 4. Neither the name of the Company nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *  
 * This software is provided ``as is'', and any express or implied
 * warranties, including, but not limited to, the implied warranties of
 * merchantability and fitness for a particular purpose are disclaimed.
 * In no event shall the company or contributors be liable for any
 * direct, indirect, incidental, special, exemplary, or consequential
 * damages (including, but not limited to, procurement of substitute
 * goods or services; loss of use, data, or profits; or business
 * interruption) however caused and on any theory of liability, whether
 * in contract, strict liability, or tort (including negligence or
 * otherwise) arising in any way out of the use of this software, even if
 * advised of the possibility of such damage.
 *
 * $Id: rawio.c,v 1.7 1999/07/21 02:10:09 grog Exp grog $
 */

#define LINELENGTH 200					    /* max length of program line */
#define MAXLINES 54					    /* number of lines on page */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <unistd.h>
#include <sys/disklabel.h>
#include <sys/errno.h>
#include <sys/param.h>
#include <sys/wait.h>
#include <sys/resource.h>
#include <signal.h>

/* submitted by Andy Doran <ad@psn.ie> */
#ifdef __NetBSD__
#include <errno.h>
#define srandomdev() srand ((unsigned) time (NULL))
#endif

#define SKIPSTART 16384					    /* amount to leave untouched at start of dev */

char buf [MAXPHYS];
int file;
int length;

int count;
enum operation
{
  RandomRead = 1,
  SequentialRead = 2,
  RandomWrite = 4,
  SequentialWrite = 8,
  }
op;
int openflags;						    /* flags to pass to open(2) */
enum operation thisop;					    /* current one we're looking at */
int test;						    /* number of test */
char *testname [] = {"RR", "SR", "RW", "SW"};		    /* names to identify tests */

char *args;
off_t offset;
int lines;
#define MAXCHILD 256					    /* maximum number of processes to start */
pid_t apid [MAXCHILD];
int nproc = 8;						    /* (default) number to start */

struct timeval start_time;
struct timeval end_time [MAXCHILD];			    /* one end time per process */
float elapsed_time;
float user_time;
float system_time;

struct rusage rusage [MAXCHILD];			    /* and associated rusage info */
struct rusage total_rusage;				    /* total values */

pid_t endproc [MAXCHILD];				    /* note the pid of each proc as it finishes */

/* Shared memory segment for communicating with
 * the children.  Each child accesses only its
 * element of the array.  The parent doesn't
 * touch them until they exit */
struct childinfo
{
  int reads;
  int writes;
  long long bytes_read;
  long long bytes_written;
  }
*childinfo;

int total_reads;
int total_writes;
long long total_bytes_read;
long long total_bytes_written;

/* child code */
void dochild (int test, int proc);

#define PAGELENGTH 60					    /* get this from termcap */
void
  checklines (int increment)
{
  char prompt [2];
  if ((lines += increment) >= PAGELENGTH)
    {
    fprintf (stderr, "More...");
    fgets (prompt, 2, stdin);
    lines = 0;
    }
  }

int maxchunk = 16384;					    /* maximum size of transfer */
long long filesize = 0;
int recs = 16384;					    /* number of records to transfer */
char *device;						    /* name of device to open */
int proc;						    /* index of this process in childinfo (child) */
							    /* index of current child (parent) */
int verbose = 0;					    /* verbosity flag */
int heading = 1;					    /* print heading if != 0 */
char *id = "anon";					    /* name of test */

void
  usage (char *argv0)
{
  fprintf (stderr,
	   "Usage: %s [options] device\n"
	   "\n"
	   "Options may be:\n"
	   "\n"
	   "\t-a\tPerform all tests (read and write, random and sequential)\n"
	   "\t-c\tMaximum I/O transfer size in sectors or bytes.\n"
	   "\t\t\tMust be an integral number of sectors, default %d bytes\n"
	   "\t-h\tSuppress heading\n"
	   "\t-I <name>\tIdentification for output\n"
	   "\t-n\tTotal number of records to write, defaults to %d\n"
	   "\t-p\tNumber of processes to start, defaults to %d\n"
	   "\t-R\tPerform random read test\n"
	   "\t-r\tPerform sequential read test\n"
	   "\t-s\tSize of file system, no default\n"
	   "\t-v\tProduce verbose output\n"
	   "\t-W\tPerform random write test\n"
	   "\t-w\tPerform sequential write test\n"
	   "\n",
	   argv0,
	   maxchunk,
	   recs,
	   nproc);
  exit (1);
  }

int main (int argc, unsigned char *argv [])
{
  int i;

  if (argc < 2)
    usage (argv [0]);
  for (i = 1; i < argc; i++)
    {
    if (argv [i][0] == '-')
      {
      switch (argv [i][1])
	{
      case 'a':						    /* do it all */
	op = SequentialRead | RandomRead | SequentialWrite | RandomWrite;
	break;
	
      case 'c':						    /* maxchunk */
	maxchunk = atoi (argv [++i]);
	if (maxchunk < 512)
	  {
	  if (maxchunk <= 0)
	    {
	    fprintf (stderr, "I/O transfer max must be at least 1 sector\n");
	    exit (1);
	    }
	  else
	    maxchunk *= 512;
	  }
	if (maxchunk > MAXPHYS)
	  {
	  printf ("Maximum transfer size limited to %d\n", MAXPHYS);
	  maxchunk = MAXPHYS;
	  }
	break;

      case 'h':
	heading = 0;
	break;

      case 'I':
	id = argv [++i];
	break;

      case 'n':
	recs = atoi (argv [++i]);
	break;

      case 'p':
	nproc = atoi (argv [++i]);
	if (nproc > MAXCHILD)				    /* too many */
	  {
	  fprintf (stderr,
		   "Can't start %d processes, maximum is %d\n",
		   nproc,
		   MAXCHILD );
	  nproc = MAXCHILD;
	  }
	break;

      case 'R':
	op |= RandomRead;
	break;

      case 'r':
	op |= SequentialRead;
	break;

      case 's':
	sscanf (argv [++i], "%qd", &filesize);
	break;

      case 'v':
	verbose = atoi (argv [++i]);
	break;

      case 'W':
	op |= RandomWrite;
	break;

      case 'w':
	op |= SequentialWrite;
	break;

      default:
	usage (argv [0]);
	}
      }
    else if (device != NULL)
      usage (argv [0]);
    else
      device = argv [i];
    }
  if (op == 0)						    /* no ops specified, */
    op = RandomRead | SequentialRead;			    /* default to the read tests */
  if (maxchunk & 0x1ff)
    {
    fprintf (stderr, "Invalid transfer size, must be multiple of 512: %d\n", maxchunk);
    exit (1);
    }
  if (device == NULL)
    {
    fprintf (stderr, "No file name specified\n");
    exit (1);
    }
  if (op & (SequentialWrite | RandomWrite) == 0)	    /* only read tests */
    openflags = O_RDONLY;				    /* open read-only */
  else
    openflags = O_RDWR;					    /* open read-write */
  if ((file = open (device, openflags, 0)) < 0)
    {
    fprintf (stderr, "Can't open %s: %s\n", device, strerror (errno));
    exit (1);
    }
  if (filesize == 0)
    {
    struct stat filestat;
    
    if (stat (device, &filestat) < 0)
      {
      fprintf (stderr, "Can't stat %s: %s\n", device, strerror (errno));
      exit (1);
      }
    else
      filesize = filestat.st_size;
    /* stat didn't do it; try lseek */
    if (filesize == 0)
      {
      filesize = lseek (file, 0, SEEK_END);		    /* try to get the end */
      if (filesize < 0)
	fprintf (stderr, "Can't seek %s: %s\n", device, strerror (errno));
      }
    /* lseek didn't do it; look at the disk label */
    if (filesize == 0)
      {
      struct disklabel label;
      if (ioctl (file, DIOCGDINFO, &label) != -1)	    /* got the label */
	filesize = ((long long) label.d_secperunit) * DEV_BSIZE;
      else if (verbose)
	fprintf (stderr, "Can't get label of %s: %s\n", device, strerror (errno));	
      }
    if (filesize == 0)
      {
      fprintf (stderr, "No file size specified\n");
      exit (1);
      }
    }
  childinfo = mmap (NULL,
		    nproc * sizeof (struct childinfo),
		    PROT_READ | PROT_WRITE,
		    MAP_INHERIT | MAP_SHARED | MAP_ANON,
		    -1,
		    0 );
  if (childinfo == MAP_FAILED)
    {
    fprintf (stderr, "Can't mmap shared memory: %s\n", strerror (errno));
    exit (1);
    }

  /* OK, we're ready to go */
  if (heading)
    {
    if (verbose > 1)
      {
      printf ("Test name: \t%16s\n"
	      "Transfer count:\t%16d\n"
	      "Record count:\t\t%8d\n"
	      "Process count:\t%16d\n"
	      "Device size:\t%16lld\n",
	      id,
	      maxchunk,
	      recs,
	      nproc,
	      filesize );
	/*     printf ("Test    ID\t\tTime\t    KB/sec\t %%User\t  %%Sys\t%%Total\tReads\tI/O in\tWrites\tI/O out\n"); */
      printf ("Test    ID\t\tTime\t    KB/sec\t    /sec %%User\t  %%Sys\t%%Total\tReads\tWrites\n");
      }
    else if (verbose)
      printf ("Test    ID\t     K/sec\t    /sec %%User\t  %%Sys\t%%Total\n");
    else
      printf ("\t   Random read\tSequential read\t   Random write\tSequential write\n"
	      "ID\t    K/sec  /sec\t   K/sec  /sec\t   K/sec  /sec\t   K/sec  /sec\n");
    }
  if (! verbose)					    /* print the ID now */
    printf ("%-8s ", id);	      

  for (test = 0; test < 4; test++)			    /* check for each test */
    {
    if (op & 1 << test)					    /* do this test */
      {
      fflush (stdout);					    /* to stop the children doing it */
      /* First, start up the procs */
      for (proc = 0; proc < nproc; proc++)		    /* spawn some processes */
	{
	pid_t pid = fork ();
	if (pid == 0)
	  dochild (test, proc);
	else if (pid < 0)
	  {
	  fprintf (stderr, "Can't fork process %d: %s\n", proc, strerror (errno));
	  exit (1);
	  }
	else
	  apid [proc] = pid;
	}
      sleep (1);					    /* give them a sec to get their act together */
      
      /* Now initialize the totals */
      bzero (&total_rusage, sizeof (struct rusage));
      total_reads = 0;
      total_writes = 0;
      total_bytes_read = 0;
      total_bytes_written = 0;
      /* Now we're ready to start: note the time */
      gettimeofday (&start_time, NULL);
      /* And let them go */
      for (proc = 0; proc < nproc; proc++)
	{
	if (kill (apid [proc], SIGUSR1) < 0)
	  fprintf (stderr, "Can't kill process %d: %s\n", apid [proc], strerror (errno));
	}
      
      /* Now wait for them to finish */
      for (proc = 0; proc < nproc; proc++)
	{
	int status;
	pid_t pid;
	
	pid = wait4 (-1, &status, 0, &rusage [proc]);
	gettimeofday (&end_time [proc], NULL);
	if (pid < 0)
	  {
	  fprintf (stderr, "Can't wait4: %s\n", strerror (errno));
	  exit (1);
	  }
	endproc [proc] = pid;				    /* note the pid */
	}
      
      for (proc = 0; proc < nproc; proc++)
	{
	total_rusage.ru_utime.tv_sec += rusage [proc].ru_utime.tv_sec; /* accumulate totals */
	total_rusage.ru_utime.tv_usec += rusage [proc].ru_utime.tv_usec; /* accumulate totals */
	total_rusage.ru_stime.tv_sec += rusage [proc].ru_stime.tv_sec; /* accumulate totals */
	total_rusage.ru_stime.tv_usec += rusage [proc].ru_stime.tv_usec; /* accumulate totals */
	total_rusage.ru_minflt += rusage [proc].ru_minflt;  /* page reclaims */
	total_rusage.ru_majflt += rusage [proc].ru_majflt;  /* page faults */
	total_rusage.ru_nswap += rusage [proc].ru_nswap;    /* swaps */
	total_rusage.ru_inblock += rusage [proc].ru_inblock; /* block input operations */
	total_rusage.ru_oublock += rusage [proc].ru_oublock; /* block output operations */
	total_rusage.ru_msgsnd += rusage [proc].ru_msgsnd;  /* messages sent */
	total_rusage.ru_msgrcv += rusage [proc].ru_msgrcv;  /* messages received */
	total_rusage.ru_nsignals += rusage [proc].ru_nsignals; /* signals received */
	total_rusage.ru_nvcsw += rusage [proc].ru_nvcsw;    /* voluntary context switches */
	total_rusage.ru_nivcsw += rusage [proc].ru_nivcsw;  /* involuntary " */
	total_reads += childinfo [proc].reads;
	total_writes += childinfo [proc].writes;
	total_bytes_read += childinfo [proc].bytes_read;
	total_bytes_written += childinfo [proc].bytes_written;
	}
      
      /* Reaped all children.  Print results */
      proc = nproc - 1;					    /* look at the last child to complete */
      elapsed_time = (end_time [proc].tv_sec - start_time.tv_sec)
	+ ((float) (end_time [proc].tv_usec - start_time.tv_usec) / 1000000);
      user_time = total_rusage.ru_utime.tv_sec + ((float) total_rusage.ru_utime.tv_usec) / 1000000;
      system_time = total_rusage.ru_stime.tv_sec + ((float) total_rusage.ru_stime.tv_usec) / 1000000;
      if (verbose > 1)
	printf ("%s %8s\t%12.6f\t%10.1f\t  %5.0f %6.1f\t%6.1f\t%6.1f\t%d\t%d\n",
		testname [test],
		id,
		elapsed_time,
		(float) (total_bytes_read + total_bytes_written) / elapsed_time / 1000,
		((float) (total_reads + total_writes) / elapsed_time),
		user_time / elapsed_time * 100,
		system_time / elapsed_time * 100,
		(user_time + system_time) / elapsed_time * 100,
		total_reads,
		total_writes );
      else if (verbose)
	{
	if (test < 2)					    /* read ops, */
	  printf ("%s %8s\t%10.1f\t  %5.0f %6.1f\t%6.1f\t%6.1f\t%d\n",
		  testname [test],
		  id,
		  (float) (total_bytes_read + total_bytes_written) / elapsed_time / 1000,
		  ((float) (total_reads + total_writes) / elapsed_time),
		  user_time / elapsed_time * 100,
		  system_time / elapsed_time * 100,
		  (user_time + system_time) / elapsed_time * 100,
		  total_reads );
	else
	  printf ("%s %8s\t%10.1f\t  %5.0f %6.1f\t%6.1f\t%6.1f\t%d\n",
		  testname [test],
		  id,
		  (float) (total_bytes_read + total_bytes_written) / elapsed_time / 1000,
		  ((float) (total_reads + total_writes) / elapsed_time),
		  user_time / elapsed_time * 100,
		  system_time / elapsed_time * 100,
		  (user_time + system_time) / elapsed_time * 100,
		  total_writes );
	}
      else						    /* not verbose, */
	printf ("%8.1f %5.0f\t",
		(float) (total_bytes_read + total_bytes_written) / elapsed_time / 1000,
		((float) (total_reads + total_writes) / elapsed_time) );
      }
    else
      printf ("\t\t");					    /* skip the column */
    }
  if (! verbose)
    printf ("\n");
  return 0;
  }

void sigcatch (int signo)
{
  }

/* Child code */
void dochild (int test, int proc)
{
  sigset_t allsigs;
  struct sigaction ignore = {sigcatch, allsigs, 0};
  int mycount;
  int i;
  int maxrecs;

  /* first, calculate the number of records to transfer.
   * If we're one of the first <remainder> processes,
   * transfer one more */
  mycount = recs / nproc;				    /* quotient */
  if (recs % nproc > proc)
    mycount++;

  /* Initialize the counters */
  childinfo [proc].reads = 0;
  childinfo [proc].bytes_read = 0;
  childinfo [proc].writes = 0;
  childinfo [proc].bytes_written = 0;

  /* Don't jump the gun */
  sigemptyset (&allsigs);
  if (sigaction (SIGUSR1, &ignore, NULL) < 0)
    {
    perror ("Can't sigaction");
    exit (1);
    }
  sigsuspend (&allsigs);

  /* OK, we're allowed to start now */
  switch (test)
    {
  case 0:						    /* random read */
    srandomdev ();
    for (i = 1; i <= mycount; i++)
      {
      int iocount;
      length = (random () % (maxchunk * 2)) & ~0x1ff;	    /* length of this transfer */
      offset = ((random () % filesize) & ~0x1ff) + SKIPSTART; /* and offset */
      if (lseek (file, offset, SEEK_SET) < 0)
	fprintf (stderr, "Child %d can't seek to %llx: %s\n", proc, offset, strerror (errno));
      else if ((iocount = read (file, buf, length)) != length)
	{
	int Errno = errno;
	off_t pos = lseek (file, 0, SEEK_CUR);
	
	if (iocount < 0)				    /* error */
	  fprintf (stderr,
		   "Child %d Bad read at %d: %s (%d), iocount %d\n",
		   proc,
		   pos,
		   strerror (Errno),
		   Errno,
		   iocount );
	else
	  fprintf (stderr,
		   "Child %d Bad read at %d: %d of %d bytes\n",
		   proc,
		   pos,
		   iocount,
		   length );
	exit (1);
	}
      childinfo [proc].reads++;
      childinfo [proc].bytes_read += iocount;
      }
    break;

  case 1:						    /*  Sequential Read */
    maxrecs = mycount;
    if (maxrecs * length > filesize - SKIPSTART)	    /* we'd overshoot, */
      maxrecs = filesize / length - SKIPSTART;		    /* trim */
    length = maxchunk;					    /* length of this transfer */
    if (lseek (file, SKIPSTART, SEEK_SET) < 0)
      fprintf (stderr, "Child %d can't seek to %llx: %s\n", proc, offset, strerror (errno));
    for (i = 1; i <= maxrecs; i++)
      {
      int iocount;
      if ((iocount = read (file, buf, length)) != length)
	{
	int Errno = errno;
	off_t pos = lseek (file, 0, SEEK_CUR);
	
	if (iocount < 0)				    /* error */
	  fprintf (stderr,
		   "Child %d Bad read at %d: %s (%d)\n",
		   proc,
		   pos,
		   strerror (Errno),
		   Errno );
	else
	  fprintf (stderr,
		   "Child %d Bad read at %d: %d of %d bytes\n",
		   proc,
		   pos,
		   iocount,
		   length );
	exit (1);
	}
      childinfo [proc].reads++;
      childinfo [proc].bytes_read += iocount;
      }
    break;

  case 2:						    /* Random Write */
    srandomdev ();
    for (i = 1; i <= mycount; i++)
      {
      int iocount;
      length = (random () % (maxchunk * 2)) & ~0x1ff;	    /* length of this transfer */
      offset = ((random () % filesize) & ~0x1ff) + SKIPSTART; /* and offset */
      if (lseek (file, offset, SEEK_SET) < 0)
	fprintf (stderr, "Child %d can't seek: %s", proc, strerror (errno));
      else if ((iocount = write (file, buf, length)) != length)
	{
	int Errno = errno;
	off_t pos = lseek (file, 0, SEEK_CUR);
	
	if (iocount < 0)				    /* error */
	  fprintf (stderr,
		   "Child %d Bad write at %d: %s (%d)\n",
		   proc,
		   pos,
		   strerror (Errno),
		   Errno );
	else
	  fprintf (stderr,
		   "Child %d Bad write at %d: %d of %d bytes\n",
		   proc,
		   pos,
		   iocount,
		   length );
	exit (1);
	}
      childinfo [proc].writes++;
      childinfo [proc].bytes_written += iocount;
      }
    break;

  case 3:						    /* Sequential Write */
    maxrecs = mycount;
    if (maxrecs * length > filesize - SKIPSTART)	    /* we'd overshoot, */
      maxrecs = filesize / length - SKIPSTART;		    /* trim */
    length = maxchunk;					    /* length of this transfer */
    if (lseek (file, SKIPSTART, SEEK_SET) < 0)		    /* back to the start */
      fprintf (stderr, "Child %d can't seek: %s", proc, strerror (errno));
    for (i = 1; i <= maxrecs; i++)
      {
      int iocount;
      if ((iocount = write (file, buf, length)) != length)
	{
	int Errno = errno;
	off_t pos = lseek (file, 0, SEEK_CUR);
	
	if (iocount < 0)				    /* error */
	  fprintf (stderr,
		   "Child %d Bad write at %d: %s (%d)\n",
		   proc,
		   pos,
		   strerror (Errno),
		   Errno );
	else
	  fprintf (stderr,
		   "Child %d Bad write at %d: %d of %d bytes\n",
		   proc,
		   pos,
		   iocount,
		   length );
	exit (1);
	}
      childinfo [proc].writes++;
      childinfo [proc].bytes_written += iocount;
      }
    break;
    }
  if (verbose > 3)					    /* really just debug */
    fprintf (stderr,
	     "Child %d: %8d %8lld %8d %8lld\n",
	     proc,
	     childinfo [proc].reads,
	     childinfo [proc].bytes_read,
	     childinfo [proc].writes,
	     childinfo [proc].bytes_written );
  exit (0);
  }
